# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Montgomery DataModule for TB detection.

Database reference: [MONTGOMERY-SHENZHEN-2014]_
"""

import os
import pathlib
import typing

import PIL.Image
from torchvision import tv_tensors
from torchvision.transforms.functional import to_tensor

from .....data.datamodule import CachingDataModule
from .....data.image_utils import remove_black_borders
from .....data.split import make_split
from .....utils.rc import load_rc
from ....data.typing import ClassificationRawDataLoader as _ClassificationRawDataLoader
from ....data.typing import Sample

CONFIGURATION_KEY_DATADIR = "datadir." + (__name__.rsplit(".", 2)[-2])
"""Key to search for in the configuration file for the root directory of this
database."""


class ClassificationRawDataLoader(_ClassificationRawDataLoader):
    """A specialized raw-data-loader for the Montgomery dataset.

    Parameters
    ----------
    config_variable
        Key to search for in the configuration file for the root directory of
        this database.
    """

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    # config_variable: required so this loader can be used for the small
    # version of the Montgomery database as well.
    def __init__(self, config_variable: str = CONFIGURATION_KEY_DATADIR):
        self.datadir = pathlib.Path(
            load_rc().get(config_variable, os.path.realpath(os.curdir)),
        )

    def sample(self, sample: tuple[str, int, typing.Any | None]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        # N.B.: Montgomery images are encoded as grayscale PNGs, so no need to
        # convert them again with Image.convert("L").
        image = PIL.Image.open(self.datadir / sample[0])
        image, _ = remove_black_borders(image)
        image = tv_tensors.Image(to_tensor(image))

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return dict(image=image), dict(target=sample[1], name=sample[0])  # type: ignore[arg-type]

    def target(self, k: typing.Any) -> int | list[int]:
        """Load a single image sample target from the disk.

        Parameters
        ----------
        k
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
        int
            The integer target associated with the sample.
        """

        return k[1]


class DataModule(CachingDataModule):
    """Montgomery DataModule for TB detection.

    The standard digital image database for Tuberculosis was created by the
    National Library of Medicine, Maryland, USA in collaboration with Shenzhen
    No.3 People’s Hospital, Guangdong Medical College, Shenzhen, China. The
    Chest X-rays are from

    * Database reference: [MONTGOMERY-SHENZHEN-2014]_
    * Original resolution (height x width or width x height): 4020x4892 px or
      4892x4020 px

    Data specifications:

    * Raw data input (on disk):

      * PNG images 8 bit grayscale
      * resolution: fixed to one of the cases above

    * Output image:

      * Transforms:

        * Load raw PNG with :py:mod:`PIL`
        * Remove black borders
        * Convert to torch tensor
        * Torch center cropping to get square image

      * Final specifications

        * Grayscale, encoded as a single plane tensor, 32-bit floats,
          square at 4020 x 4020 pixels
        * Labels: 0 (healthy), 1 (active tuberculosis)

    Parameters
    ----------
    split_filename
        Name of the .json file containing the split to load.
    """

    def __init__(self, split_filename: str):
        assert __package__ is not None
        super().__init__(
            database_split=make_split(__package__, split_filename),
            raw_data_loader=ClassificationRawDataLoader(),
            database_name=__package__.rsplit(".", 1)[1],
            split_name=pathlib.Path(split_filename).stem,
            task="classification",
        )
